/***********************************************************************/
/*                                                                     */
/* RIGI Utilities --- Copyright (C) 1996-1999                          */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Author: Johannes Martin (jmartin@csr.csc.uvic.ca)                   */
/*                                                                     */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Parameters passed in by the command line switchs */
enum ArgType {
  ArgNone, ArgError, ArgHelp, ArgS = 8, ArgP = 16, ArgT = 32, ArgX = 64
};

static int ParseCmdLine(int argc, char **argv, CS_String& project) {
  int i, j, k, args = 0;
  int isLast;
  char c;

  for (i = 1; i < argc; i++) {
    if (*argv[i] != '-') {
      if (project == "") {
        project = argv[i];
        continue;
      } else {
        fprintf(stderr, "Invalid argument \"%s\"\n", argv[i]);
        return (ArgError | ArgHelp);
      }
    }
    k = strlen(argv[i]);
    for (j = 1; j < k; j++) {
      c = *(argv[i] + j);
      isLast = !(*(argv[i] + j + 1));	/* last character in argument? */
      switch (c) {
      case 'h':
   args |= (int)ArgHelp;
	break;
      case 's':
	if (isLast && i + 1 < argc && j + 1 == k && argv[i + 1][0] != '-') {
	  maxLines = atoi(argv[++i]);
	  if (maxLines < 1) {
	    fprintf(stderr, "-l must be followed by a non-negative number of lines\n");
	    return (ArgError | ArgHelp);
	  }
	}
	args |= (int)ArgS;
	break;
      case 'p':
	args |= (int)ArgP;
	break;
      case 't':
	if (isLast && i + 1 < argc && j + 1 == k && argv[i + 1][0] != '-') {
	  args |= (int)ArgT;
	  if (!HtLoadTemplate(argv[++i])) return (ArgError | ArgHelp);
	  break;
	}
	fprintf(stderr, "-t must be followed by a template file name\n");
	return (ArgError | ArgHelp);
      case 'x':
	if (isLast && i + 1 < argc && j + 1 == k && argv[i + 1][0] != '-') {
	  tabSize = atoi(argv[++i]);
	  if (tabSize > 10 || tabSize < 1) {
	    fprintf(stderr, "-x expands tabs to a maximum of 10 spaces\n");
	    return (ArgError | ArgHelp);
	  }
	} else {
	  tabSize = 4;
	}
	args |= (int)ArgX;
	break;
      default:
	fprintf(stderr, "Invalid argument \"%s\"\n", argv[i]);
	return (ArgError | ArgHelp);
      }
    }
  }
  return args;
}

static void PrintHelp(char *progName) {
  fprintf(stderr,
	  "HTMLize Source Using IBM VA C++ Codestore (cs2html) - " VERSION "\n");
  fprintf(stderr,
	  "Usage: %s [-h] [-t template] [-s maxLines] [-x numSpaces]\n", progName);
  fprintf(stderr,
	  "\t-h\t\thelp\n");
  fprintf(stderr,
	  "\t-s maxLines\tsegment source files at a maximum size of maxLines\n");
  fprintf(stderr,
	  "\t-p\t\tapply \"preformatted\" (<pre>) HTML tags\n");
  fprintf(stderr,
	  "\t-t template\tuse HTML template file\n");
  fprintf(stderr,
	  "\t-x numSpaces\texpand tabs to spaces (default is 4 spaces)\n");
}

class CodeStore2Html : public CodeStoreParser {
  htmlTupleSet tuples;

  virtual void emit(const CS_String& elem0, const CS_String& elem1, const CS_String& elem2,
                    const CS_String& elem3, int line, int column);

public:
  void parse(CS_CodeStore &cs, bool segmented, bool preformatted) {
    CodeStoreParser::parse(cs);
    if (!Htmlize(tuples, segmented, preformatted))
      exit(1);
  }
};

void CodeStore2Html::emit(const CS_String& elem0, const CS_String& elem1, const CS_String& elem2,
                          const CS_String& elem3, int line, int column) {
  HtmlRsfTuple tuple(elem0.str(), elem1.str(), elem2.str(), elem3.str(), line, column);
  if (tuple.htmlType != HtmlRsfTuple::NoType)
    tuples.insert(tuple);
}

int main(int argc, char *argv[]) {
  CS_String project("");
  int args = ParseCmdLine(argc, argv, project);
  if ((project == "") || (args & ArgHelp)) {
    PrintHelp(argv[0]);
    return (args & ArgError);
  }
  HtmlRsfTuple::addAnchorString(UniqueString("returns"));
  HtmlRsfTuple::addAnchorString(UniqueString("hasType"));
  HtmlRsfTuple::addAnchorString(UniqueString("extends"));
  HtmlRsfTuple::addAnchorString(UniqueString("isTheSameAs"));
  HtmlRsfTuple::addAnchorString(UniqueString("calls"));
  HtmlRsfTuple::addAnchorString(UniqueString("references"));
  HtmlRsfTuple::addAnchorString(UniqueString("declares"));
  HtmlRsfTuple::addAnchorString(UniqueString("tagged"));
  HtmlRsfTuple::addBookmarkString(UniqueString("type"));

  // Create a Codestore, connect it to the sources listed in the project file.

  CS_CodeStore::ReturnValue r;
  CS_CodeStore* code_store;

  r = CS_CodeStore::createTransientDatabase(code_store,project.str());

  if (r != 0) {
    PrintCodeStoreError(r);
    cerr << endl;
    PauseBeforeQuitting();
    exit(1);
  }

  // Create a default incorporator.
  //
  CS_IncorporationController controller;

  // Customize the controller here.

  //  Load the database.
  //
  CS_bool succeeded = code_store->incorporate(controller);

  if (!succeeded) {
    CS_MessageList &l = code_store->messages();
    PrintMessages(l.first());
  }

  // Browse the CodeStore here.
  try {
    CodeStore2Html().parse(*code_store, args & ArgS, args & ArgP);
  } catch (CS_Error& e) {
    cerr << "Caught exception: " << e.message() << endl;
  } catch (...) {
    cerr << "Uncaught exception!" << endl;
  }
}
