/***********************************************************************/
/*                                                                     */
/* RIGI Utilities --- Copyright (C) 1996-1999                          */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Author: Johannes Martin (jmartin@csr.csc.uvic.ca)                   */
/*                                                                     */
/***********************************************************************/
using std::string;
using std::list;
using std::map;

class CodeStoreParser {
protected:
  virtual void emit(const CS_String& elem0, const CS_String& elem1, const CS_String& elem2,
                    const CS_String& elem3, int line, int column) = 0;
  virtual void emit(const CS_String& elem0, const CS_String& elem1, const CS_String& elem2,
                    CS_SourceLocation* sl = NULL);

public:
  virtual void emitAccesses(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitAccessKind(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitCalls(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitClassKind(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitExtends(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitHasType(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitIsCastTo(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitIsDefinedIn(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitIsInNamespace(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitIsMemberOf(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitIsTheSameAs(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitReferences(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitReturns(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitTagged(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);
  virtual void emitType(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl = NULL);

  virtual CS_String toString(CS_TypeDescriptor& td);
  bool mapsToPredefinedType(CS_TypeDescriptor& td);

  virtual void parse(CS_Implementation& impl);
  virtual void parse(CS_Declaration& d);
  virtual void parse(CS_DeclarationStore& ds);
  virtual void parse(CS_CodeStore& cs);
};

void CodeStoreParser::emit(const CS_String& elem0, const CS_String& elem1,
                           const CS_String& elem2, CS_SourceLocation* sl) {
  if (sl && !sl->isGenerated())
    emit(elem0, elem1, elem2, sl->fileName(), sl->line(), sl->column());
  else
    emit(elem0, elem1, elem2, "", 0, 0);
}

void CodeStoreParser::emitAccesses(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("accesses", src, dst, sl);
}

void CodeStoreParser::emitAccessKind(const CS_String& src, const CS_String& dst, CS_SourceLocation *sl) {
  emit("accessKind", src, dst, sl);
}

void CodeStoreParser::emitCalls(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("calls", src, dst, sl);
}

void CodeStoreParser::emitClassKind(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("classKind", src, dst, sl);
}

void CodeStoreParser::emitExtends(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("extends", src, dst, sl);
}

void CodeStoreParser::emitHasType(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("hasType", src, dst, sl);
}

void CodeStoreParser::emitIsCastTo(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("isCastTo", src, dst, sl);
}

void CodeStoreParser::emitIsDefinedIn(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("isDefinedIn", src, dst, sl);
}

void CodeStoreParser::emitIsInNamespace(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("isInNamespace", src, dst, sl);
}

void CodeStoreParser::emitIsMemberOf(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("isMemberOf", src, dst, sl);
}

void CodeStoreParser::emitIsTheSameAs(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("isTheSameAs", src, dst, sl);
}

void CodeStoreParser::emitReferences(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("references", src, dst, sl);
}

void CodeStoreParser::emitReturns(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("returns", src, dst, sl);
}

void CodeStoreParser::emitTagged(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("tagged", src, dst, sl);
}

void CodeStoreParser::emitType(const CS_String& src, const CS_String& dst, CS_SourceLocation* sl) {
  emit("type", src, dst, sl);
}

CS_String CodeStoreParser::toString(CS_TypeDescriptor& td) {
  BEGINTRY
  switch (td.typeDescriptorKind()) {
    case CS_TypeDescriptor::IsPointer:
    case CS_TypeDescriptor::IsReference:
    case CS_TypeDescriptor::IsPointerToMember:
    case CS_TypeDescriptor::IsArray:
      return toString(*td.next());
    case CS_TypeDescriptor::IsFunction:
      return td.signature("function", true, false);
    case CS_TypeDescriptor::IsPredefinedTypeSpecifier:
    case CS_TypeDescriptor::IsNamedTypeSpecifier:
    case CS_TypeDescriptor::IsUnnamedTypeSpecifier:
      return UniqueIdentifier::idForDeclaration(td.declaration());
  }
  ENDTRY
  cerr << "unknown type descriptor kind" << endl;
  return "(unknown)";
}

bool CodeStoreParser::mapsToPredefinedType(CS_TypeDescriptor& td) {
  BEGINTRY
  switch (td.typeDescriptorKind()) {
    case CS_TypeDescriptor::IsPointer:
    case CS_TypeDescriptor::IsReference:
    case CS_TypeDescriptor::IsPointerToMember:
    case CS_TypeDescriptor::IsArray:
      return mapsToPredefinedType(*td.next());
    case CS_TypeDescriptor::IsFunction:
      return false;
    case CS_TypeDescriptor::IsPredefinedTypeSpecifier:
      return true;
    case CS_TypeDescriptor::IsNamedTypeSpecifier:
    case CS_TypeDescriptor::IsUnnamedTypeSpecifier:
      return false;
  }
  ENDTRY
  cerr << "unknown type descriptor kind" << endl;
  return false;
}

void CodeStoreParser::parse(CS_Implementation& impl) {
  BEGINTRY
  CS_StorageArena arena;
  CS_StoragePlacement arenaStorage(&arena);
  switch (impl.implementationKind()) {
    case CS_Implementation::IsFunctionBody:
      {
        CS_FunctionBody *fb = impl.asFunctionBody();
        CS_FunctionBody::BodyInfo *b;

        fb->body(b, CS_Implementation::analyzed, arenaStorage);
        if (b && b->body())
          CSImplementationParser(*this, impl).visitStatement(*b->body());
      }
      break;
    case CS_Implementation::IsVariableInitializer:
      {
        CS_VariableInitializer *vi = impl.asVariableInitializer();
        CS_VariableDeclaration* var = impl.declaration()->asVariableDeclaration();
        CS_Initializer *i;
        vi->initializer(i, CS_Implementation::analyzed, arenaStorage);
        if (i)
          CSImplementationParser(*this, impl).visitInitializer(*i, &var->typeDescriptor(), var);
      }
      break;
    default:
      cerr << "Unknown implementation kind" << endl;
  }
  ENDTRY
}

void CodeStoreParser::parse(CS_Declaration& d) {
  CS_String rsfNodeType;

  BEGINTRY
  switch (d.declarationKind()) {
    case CS_Declaration::IsPredefinedType:
      rsfNodeType = "Datatype";
      break;
    case CS_Declaration::IsFunction:
      if (d.asFunctionDeclaration()->typeDescriptor().returnType())
        emitReturns(UniqueIdentifier::idForDeclaration(d),
                    toString(*d.asFunctionDeclaration()->typeDescriptor().returnType()),
                    d.aSourceLocation());
      if (d.implementation()) {
        parse(*d.implementation());
        rsfNodeType = "Function";
      } else
        rsfNodeType = "Prototype";
      break;
    case CS_Declaration::IsVariable:
      emitHasType(UniqueIdentifier::idForDeclaration(d),
                  toString(d.asVariableDeclaration()->typeDescriptor()),
                  d.aSourceLocation());
      if (d.implementation())
        parse(*d.implementation());
      if (d.asVariableDeclaration()->typeDescriptor().isConst() ||
	  d.asVariableDeclaration()->isEnumerator())
	rsfNodeType = "Constant";
      else
	rsfNodeType = "Variable";
      break;
    case CS_Declaration::IsClass:
      {
        CS_ClassDeclaration& cd = *d.asClassDeclaration();
        CS_String kind;
        switch (cd.classDeclarationKind()) {
          case CS_ClassDeclaration::IsClass:
            kind = "class";
            break;
          case CS_ClassDeclaration::IsStruct:
            kind = "struct";
            break;
          case CS_ClassDeclaration::IsUnion:
            kind = "union";
            break;
        }
        emitClassKind(UniqueIdentifier::idForDeclaration(d), kind);
        parse(cd.memberDeclarationStore());
        for (CS_BaseSpecifier* bs = cd.firstBaseSpecifier(); bs; bs = bs->next())
          emitExtends(UniqueIdentifier::idForDeclaration(d),
                      UniqueIdentifier::idForDeclaration(bs->classDeclaration()),
                      d.aSourceLocation());
        rsfNodeType = "Class";
      }
      break;
    case CS_Declaration::IsEnumeration:
      rsfNodeType = "Datatype";
      break;
    case CS_Declaration::IsTypedef:
      emitIsTheSameAs(UniqueIdentifier::idForDeclaration(d),
                      toString(d.asTypedefDeclaration()->typeDescriptor()),
                      d.aSourceLocation());
      rsfNodeType = "Datatype"; break;
    case CS_Declaration::IsFriend:
      return;
    case CS_Declaration::IsNamespace:
      rsfNodeType = "Namespace";
      break;
    case CS_Declaration::IsUsing:
      return;
    case CS_Declaration::IsUsingDirective:
      return;
    default:
      cerr << "Unknown declaration kind" << endl;
      return;
  }
  ENDTRY
  BEGINTRY
  emitType(UniqueIdentifier::idForDeclaration(d),
           rsfNodeType, d.aSourceLocation());
  ENDTRY
  BEGINTRY
  CS_DeclarationStore& ds = d.enclosingDeclarationStore();
  switch (ds.declarationStoreKind()) {
    case CS_DeclarationStore::IsGlobal:
      break;
    case CS_DeclarationStore::IsNamespace:
      emitIsInNamespace(UniqueIdentifier::idForDeclaration(d),
                        UniqueIdentifier::idForDeclaration(ds.asNamespaceDeclarationStore()->namespaceDeclaration()),
                        d.aSourceLocation());
      break;
    case CS_DeclarationStore::IsTemplate:
      break;
    case CS_DeclarationStore::IsMember:
      {
        CS_String kind;
        switch (d.accessKind()) {
          case CS_Access::IsPublic:
            kind = "public";
            break;
          case CS_Access::IsPrivate:
            kind = "private";
            break;
          case CS_Access::IsProtected:
            kind = "protected";
            break;
        }
        emitAccessKind(UniqueIdentifier::idForDeclaration(d), kind);
        emitIsMemberOf(UniqueIdentifier::idForDeclaration(d),
                       UniqueIdentifier::idForDeclaration(ds.asMemberDeclarationStore()->classDeclaration()),
                       d.aSourceLocation());
      }
      break;
    case CS_DeclarationStore::IsLexicalBlock:
      emitIsDefinedIn(UniqueIdentifier::idForDeclaration(d),
                      UniqueIdentifier::idForDeclaration(ds.asLexicalBlockDeclarationStore()->enclosingFunction()),
                      d.aSourceLocation());
      break;
    case CS_DeclarationStore::IsTarget:
      break;
    case CS_DeclarationStore::IsNameDeclarator:
      break;
    default:
      cerr << "Unknown declaration scope kind" << endl;
      return;
  }
  ENDTRY
  BEGINTRY
  if (d.aSourceLocation() && d.aSourceLocation()->isDefiningLocation()) {
    for (CS_Location* l = d.firstLocation(); l; l = l->next())
      if ((l->locationKind() == CS_Location::IsSource) && !l->isDefiningLocation())
        emitTagged(UniqueIdentifier::idForDeclaration(d),
                   UniqueIdentifier::idForDeclaration(d),
                   l->asSourceLocation());
  }
  ENDTRY
}

void CodeStoreParser::parse(CS_DeclarationStore& ds) {
  for (CS_Declaration* d = ds.firstDeclaration(); d; d = d->next())
    parse(*d);
}

// print the RSF for the entire CodeStore
void CodeStoreParser::parse(CS_CodeStore& cs) {
  parse(cs.globalDeclarationStore());
}
